@@ -13,6 +13,8 @@ import (
 	"sync"
 	"time"
 
+	"github.com/openshift/oc/pkg/cli/admin/inspect"
+
 	"github.com/spf13/cobra"
 
 	corev1 "k8s.io/api/core/v1"
@@ -31,6 +33,8 @@ import (
 	"k8s.io/kubectl/pkg/scheme"
 	"k8s.io/kubectl/pkg/util/templates"
 
+	imagereference "github.com/openshift/library-go/pkg/image/reference"
+
 	imagev1client "github.com/openshift/client-go/image/clientset/versioned/typed/image/v1"
 	"github.com/openshift/library-go/pkg/image/imageutil"
 	"github.com/openshift/library-go/pkg/operator/resource/retry"
@@ -55,6 +59,9 @@ var (
 		# gather information with a specific local folder to copy to
 		  oc adm must-gather --dest-dir=/local/directory
 
+		# gather audit information
+		  oc adm must-gather -- /usr/bin/gather_audit_logs
+
 		# gather information using multiple plug-in images
 		  oc adm must-gather --image=quay.io/kubevirt/must-gather --image=quay.io/openshift/origin-must-gather
 
@@ -226,7 +233,7 @@ func (o *MustGatherOptions) Validate() error {
 func (o *MustGatherOptions) Run() error {
 	var err error
 
-	// create namespace
+	// create namespace ...
 	ns, err := o.Client.CoreV1().Namespaces().Create(context.TODO(), &corev1.Namespace{
 		ObjectMeta: metav1.ObjectMeta{
 			GenerateName: "openshift-must-gather-",
@@ -234,7 +241,8 @@ func (o *MustGatherOptions) Run() error {
 				"openshift.io/run-level": "0",
 			},
 			Annotations: map[string]string{
-				"oc.openshift.io/command": "oc adm must-gather",
+				"oc.openshift.io/command":    "oc adm must-gather",
+				"openshift.io/node-selector": "",
 			},
 		},
 	}, metav1.CreateOptions{})
@@ -245,13 +253,14 @@ func (o *MustGatherOptions) Run() error {
 	if !o.Keep {
 		defer func() {
 			if err := o.Client.CoreV1().Namespaces().Delete(context.TODO(), ns.Name, metav1.DeleteOptions{}); err != nil {
-				fmt.Printf("%v", err)
+				fmt.Printf("%v\n", err)
 				return
 			}
 			o.PrinterDeleted.PrintObj(ns, o.LogOut)
 		}()
 	}
 
+	// ... cluster role binding ...
 	clusterRoleBinding, err := o.Client.RbacV1().ClusterRoleBindings().Create(context.TODO(), o.newClusterRoleBinding(ns.Name), metav1.CreateOptions{})
 	if err != nil {
 		return err
@@ -260,16 +269,22 @@ func (o *MustGatherOptions) Run() error {
 	if !o.Keep {
 		defer func() {
 			if err := o.Client.RbacV1().ClusterRoleBindings().Delete(context.TODO(), clusterRoleBinding.Name, metav1.DeleteOptions{}); err != nil {
-				fmt.Printf("%v", err)
+				fmt.Printf("%v\n", err)
 				return
 			}
 			o.PrinterDeleted.PrintObj(clusterRoleBinding, o.LogOut)
 		}()
 	}
 
-	// create pods
+	// ... and finally must-gather pod
 	var pods []*corev1.Pod
 	for _, image := range o.Images {
+		_, err := imagereference.Parse(image)
+		if err != nil {
+			o.log("unable to parse image reference %s: %v", image, err)
+			return err
+		}
+
 		pod, err := o.Client.CoreV1().Pods(ns.Name).Create(context.TODO(), o.newPod(o.NodeName, image), metav1.CreateOptions{})
 		if err != nil {
 			return err
@@ -278,9 +293,18 @@ func (o *MustGatherOptions) Run() error {
 		pods = append(pods, pod)
 	}
 
+	// log timestamps...
+	if err := os.MkdirAll(o.DestDir, os.ModePerm); err != nil {
+		return err
+	}
+	if err := o.logTimestamp(); err != nil {
+		return err
+	}
+	defer o.logTimestamp()
+
 	var wg sync.WaitGroup
 	wg.Add(len(pods))
-	errs := make(chan error, len(pods))
+	errCh := make(chan error, len(pods))
 	for _, pod := range pods {
 		go func(pod *corev1.Pod) {
 			defer wg.Done()
@@ -290,7 +314,7 @@ func (o *MustGatherOptions) Run() error {
 			// wait for gather container to be running (gather is running)
 			if err := o.waitForGatherContainerRunning(pod); err != nil {
 				log("gather did not start: %s", err)
-				errs <- fmt.Errorf("gather did not start for pod %s: %s", pod.Name, err)
+				errCh <- fmt.Errorf("gather did not start for pod %s: %s", pod.Name, err)
 				return
 			}
 			// stream gather container logs
@@ -300,9 +324,9 @@ func (o *MustGatherOptions) Run() error {
 
 			// wait for pod to be running (gather has completed)
 			log("waiting for gather to complete")
-			if err := o.waitForPodRunning(pod); err != nil {
+			if err := o.waitForGatherToComplete(pod); err != nil {
 				log("gather never finished: %v", err)
-				errs <- fmt.Errorf("gather never finished for pod %s: %s", pod.Name, err)
+				errCh <- fmt.Errorf("gather never finished for pod %s: %s", pod.Name, err)
 				return
 			}
 
@@ -311,23 +335,29 @@ func (o *MustGatherOptions) Run() error {
 			pod, err = o.Client.CoreV1().Pods(pod.Namespace).Get(context.TODO(), pod.Name, metav1.GetOptions{})
 			if err != nil {
 				log("gather output not downloaded: %v\n", err)
-				errs <- fmt.Errorf("unable to download output from pod %s: %s", pod.Name, err)
+				errCh <- fmt.Errorf("unable to download output from pod %s: %s", pod.Name, err)
 				return
 			}
 			if err := o.copyFilesFromPod(pod); err != nil {
 				log("gather output not downloaded: %v\n", err)
-				errs <- fmt.Errorf("unable to download output from pod %s: %s", pod.Name, err)
+				errCh <- fmt.Errorf("unable to download output from pod %s: %s", pod.Name, err)
 				return
 			}
 		}(pod)
 	}
 	wg.Wait()
-	close(errs)
-	var arr []error
-	for i := range errs {
-		arr = append(arr, i)
+	close(errCh)
+	var errs []error
+	for i := range errCh {
+		errs = append(errs, i)
 	}
-	return errors.NewAggregate(arr)
+
+	// now gather all the events into a single file and produce a unified file
+	if err := inspect.CreateEventFilterPage(o.DestDir); err != nil {
+		errs = append(errs, err)
+	}
+
+	return errors.NewAggregate(errs)
 }
 
 func newPodOutLogger(out io.Writer, podName string) func(string, ...interface{}) {
@@ -341,11 +371,20 @@ func (o *MustGatherOptions) log(format string, a ...interface{}) {
 	fmt.Fprintf(o.LogOut, format+"\n", a...)
 }
 
+func (o *MustGatherOptions) logTimestamp() error {
+	f, err := os.OpenFile(path.Join(o.DestDir, "timestamp"), os.O_WRONLY|os.O_CREATE|os.O_APPEND, 0644)
+	if err != nil {
+		return err
+	}
+	_, err = f.WriteString(fmt.Sprintf("%v\n", time.Now()))
+	return err
+}
+
 func (o *MustGatherOptions) copyFilesFromPod(pod *corev1.Pod) error {
 	streams := o.IOStreams
 	streams.Out = newPrefixWriter(streams.Out, fmt.Sprintf("[%s] OUT", pod.Name))
-	destDir := path.Join(o.DestDir, regexp.MustCompile("[^A-Za-z0-9]+").ReplaceAllString(pod.Status.InitContainerStatuses[0].ImageID, "-"))
-	if err := os.MkdirAll(destDir, 0775); err != nil {
+	destDir := path.Join(o.DestDir, regexp.MustCompile("[^A-Za-z0-9]+").ReplaceAllString(pod.Status.ContainerStatuses[0].ImageID, "-"))
+	if err := os.MkdirAll(destDir, os.ModePerm); err != nil {
 		return err
 	}
 	rsyncOptions := &rsync.RsyncOptions{
@@ -355,7 +394,7 @@ func (o *MustGatherOptions) copyFilesFromPod(pod *corev1.Pod) error {
 		Destination:   &rsync.PathSpec{PodName: "", Path: destDir},
 		Client:        o.Client,
 		Config:        o.Config,
-		RshCmd:        fmt.Sprintf("%s --namespace=%s", o.RsyncRshCmd, pod.Namespace),
+		RshCmd:        fmt.Sprintf("%s --namespace=%s -c copy", o.RsyncRshCmd, pod.Namespace),
 		IOStreams:     streams,
 	}
 	rsyncOptions.Strategy = rsync.NewDefaultCopyStrategy(rsyncOptions)
@@ -368,7 +407,7 @@ func (o *MustGatherOptions) getGatherContainerLogs(pod *corev1.Pod) error {
 		ResourceArg: pod.Name,
 		Options: &corev1.PodLogOptions{
 			Follow:    true,
-			Container: pod.Spec.InitContainers[0].Name,
+			Container: pod.Spec.Containers[0].Name,
 		},
 		RESTClientGetter: o.RESTClientGetter,
 		Object:           pod,
@@ -389,9 +428,8 @@ func newPrefixWriter(out io.Writer, prefix string) io.Writer {
 	return writer
 }
 
-func (o *MustGatherOptions) waitForPodRunning(pod *corev1.Pod) error {
-	phase := pod.Status.Phase
-	err := wait.PollImmediate(time.Duration(o.Timeout)*time.Second, time.Minute, func() (bool, error) {
+func (o *MustGatherOptions) waitForGatherToComplete(pod *corev1.Pod) error {
+	err := wait.PollImmediate(10*time.Second, time.Duration(o.Timeout)*time.Second, func() (bool, error) {
 		var err error
 		if pod, err = o.Client.CoreV1().Pods(pod.Namespace).Get(context.TODO(), pod.Name, metav1.GetOptions{}); err != nil {
 			// at this stage pod should exist, we've been gathering container logs, so error if not found
@@ -400,15 +438,26 @@ func (o *MustGatherOptions) waitForPodRunning(pod *corev1.Pod) error {
 			}
 			return false, nil
 		}
-		phase = pod.Status.Phase
-		switch phase {
-		case corev1.PodRunning:
-			return true, nil
-		case corev1.PodFailed, corev1.PodSucceeded:
-			return true, fmt.Errorf("%s/%s unexpected pod phase: %s", pod.Namespace, pod.Name, phase)
-		default:
+		var state *corev1.ContainerState
+		for _, cstate := range pod.Status.ContainerStatuses {
+			if cstate.Name == "gather" {
+				state = &cstate.State
+				break
+			}
+		}
+
+		// missing status for gather container => timeout in the worst case
+		if state == nil {
 			return false, nil
 		}
+
+		if state.Terminated != nil {
+			if state.Terminated.ExitCode == 0 {
+				return true, nil
+			}
+			return true, fmt.Errorf("%s/%s unexpectedly terminated: exit code: %v, reason: %s, message: %s", pod.Namespace, pod.Name, state.Terminated.ExitCode, state.Terminated.Reason, state.Terminated.Message)
+		}
+		return false, nil
 	})
 	if err != nil {
 		return err
@@ -417,15 +466,18 @@ func (o *MustGatherOptions) waitForPodRunning(pod *corev1.Pod) error {
 }
 
 func (o *MustGatherOptions) waitForGatherContainerRunning(pod *corev1.Pod) error {
-	return wait.PollImmediate(time.Duration(o.Timeout)*time.Second, time.Minute, func() (bool, error) {
+	return wait.PollImmediate(10*time.Second, time.Duration(o.Timeout)*time.Second, func() (bool, error) {
 		var err error
 		if pod, err = o.Client.CoreV1().Pods(pod.Namespace).Get(context.TODO(), pod.Name, metav1.GetOptions{}); err == nil {
-			if len(pod.Status.InitContainerStatuses) == 0 {
+			if len(pod.Status.ContainerStatuses) == 0 {
 				return false, nil
 			}
-			state := pod.Status.InitContainerStatuses[0].State
-			if state.Waiting != nil && state.Waiting.Reason == "ErrImagePull" {
-				return true, fmt.Errorf("unable to pull image: %v: %v", state.Waiting.Reason, state.Waiting.Message)
+			state := pod.Status.ContainerStatuses[0].State
+			if state.Waiting != nil {
+				switch state.Waiting.Reason {
+				case "ErrImagePull", "ImagePullBackOff", "InvalidImageName":
+					return true, fmt.Errorf("unable to pull image: %v: %v", state.Waiting.Reason, state.Waiting.Message)
+				}
 			}
 			running := state.Running != nil
 			terminated := state.Terminated != nil
@@ -466,6 +518,14 @@ func (o *MustGatherOptions) newClusterRoleBinding(ns string) *rbacv1.ClusterRole
 // - copy: no-op container we can exec into
 func (o *MustGatherOptions) newPod(node, image string) *corev1.Pod {
 	zero := int64(0)
+
+	nodeSelector := map[string]string{
+		corev1.LabelOSStable: "linux",
+	}
+	if node == "" {
+		nodeSelector["node-role.kubernetes.io/master"] = ""
+	}
+
 	ret := &corev1.Pod{
 		ObjectMeta: metav1.ObjectMeta{
 			GenerateName: "must-gather-",
@@ -484,25 +544,26 @@ func (o *MustGatherOptions) newPod(node, image string) *corev1.Pod {
 					},
 				},
 			},
-			InitContainers: []corev1.Container{
-				{
-					Name:    "gather",
-					Image:   image,
-					Command: []string{"/usr/bin/gather"},
-					VolumeMounts: []corev1.VolumeMount{
-						{
-							Name:      "must-gather-output",
-							MountPath: path.Clean(o.SourceDir),
-							ReadOnly:  false,
-						},
-					},
-				},
-			},
 			Containers: []corev1.Container{
 				{
-					Name:    "copy",
-					Image:   image,
-					Command: []string{"/bin/bash", "-c", "trap : TERM INT; sleep infinity & wait"},
+					Name:            "gather",
+					Image:           image,
+					ImagePullPolicy: corev1.PullIfNotPresent,
+					// always force disk flush to ensure that all data gathered is accessible in the copy container
+					Command: []string{"/bin/bash", "-c", "/usr/bin/gather; sync"},
+					VolumeMounts: []corev1.VolumeMount{
+						{
+							Name:      "must-gather-output",
+							MountPath: path.Clean(o.SourceDir),
+							ReadOnly:  false,
+						},
+					},
+				},
+				{
+					Name:            "copy",
+					Image:           image,
+					ImagePullPolicy: corev1.PullIfNotPresent,
+					Command:         []string{"/bin/bash", "-c", "trap : TERM INT; sleep infinity & wait"},
 					VolumeMounts: []corev1.VolumeMount{
 						{
 							Name:      "must-gather-output",
@@ -512,6 +573,7 @@ func (o *MustGatherOptions) newPod(node, image string) *corev1.Pod {
 					},
 				},
 			},
+			NodeSelector:                  nodeSelector,
 			TerminationGracePeriodSeconds: &zero,
 			Tolerations: []corev1.Toleration{
 				{
@@ -521,7 +583,9 @@ func (o *MustGatherOptions) newPod(node, image string) *corev1.Pod {
 		},
 	}
 	if len(o.Command) > 0 {
-		ret.Spec.InitContainers[0].Command = o.Command
+		// always force disk flush to ensure that all data gathered is accessible in the copy container
+		ret.Spec.Containers[0].Command = []string{"/bin/bash", "-c", fmt.Sprintf("%s; sync", strings.Join(o.Command, " "))}
 	}
+
 	return ret
 }
